
# from vwsfriend.util.location_util import locationFromLatLon

def test_locationFromLatLon():
    # location = locationFromLatLon(latitude=53.60853453781239, longitude=10.19093520255688)
    # print (location)
    # assert location is not None

    # location = locationFromLatLon(latitude=53.60957750612437, longitude=10.184564007135274)
    # print (location)
    # assert location is not None

    # location = locationFromLatLon(latitude=53.61014, longitude=10.18435)
    # print (location)
    # assert location is not None

    # location = locationFromLatLon(latitude=53.66243, longitude=10.16044)
    # print (location)
    # assert location is not None

    # location = locationFromLatLon(latitude=53.65190, longitude=10.16269)
    # print (location)
    # assert location is not None
    assert True



import threading
import os
import uuid
import logging
import flask

from flask_wtf.csrf import CSRFProtect

from werkzeug.serving import make_server

from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from vwsfriend.model.base import Base

import vwsfriend.ui.status as status
import vwsfriend.ui.settings as settings

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
DEFAULT_LOG_LEVEL = "ERROR"

LOG = logging.getLogger("VWsFriend")

csrf = CSRFProtect()


class VWsFriendUI:
    def __init__(self, weConnect=None, connector=None, dbUrl=None):
        print(os.path.dirname(__file__))
        self.app = flask.Flask('VWsFriend', template_folder=os.path.dirname(__file__) + '/templates', static_folder=os.path.dirname(__file__) + '/static')
        self.app.debug = True
        self.app.config.from_mapping(
            SECRET_KEY=uuid.uuid4().hex,
        )
        csrf.init_app(self.app)

        self.app.add_url_rule('/', '/', self.root)

        self.app.register_blueprint(status.bp)
        self.app.register_blueprint(settings.bp)
        self.app.weConnect = weConnect
        self.app.connector = connector

        if connector.withDB and dbUrl is not None:
            engine = create_engine(dbUrl)
            self.app.session = Session(engine)
            Base.metadata.create_all(engine)

    def run(self, host="0.0.0.0", port=4000, loglevel=logging.INFO):  # nosec
        os.environ['WERKZEUG_RUN_MAIN'] = 'true'
        log = logging.getLogger('werkzeug')
        log.setLevel(loglevel)

        server = make_server(host, port, self.app)

        webthread = threading.Thread(target=server.serve_forever)
        webthread.start()
        LOG.info('VWsFriend is listening on %s:%s)', host, port)

        # server.shutdown()
        # webthread.join()
        # LOG.info('ProSafeExporter was stopped')

    def root(self):
        return flask.redirect(flask.url_for('status.vehicles'))



import logging

import pyhap

from weconnect.elements.climatization_status import ClimatizationStatus
from weconnect.addressable import AddressableLeaf

LOG = logging.getLogger("VWsFriend")


class Climatization(pyhap.accessory.Accessory):
    """Climatization Assessory"""

    category = pyhap.const.CATEGORY_THERMOSTAT

    def __init__(self, driver, displayName, climatizationStatus, climatizationSettings=None):
        super().__init__(driver=driver, display_name=displayName)

        servThermostat = self.add_preload_service('Thermostat')

        if climatizationStatus.climatisationState.enabled:
            climatizationStatus.climatisationState.addObserver(self.onClimatizationState,
                                                               AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            self.charCurrentHeatingCoolingState = servThermostat.configure_char('CurrentHeatingCoolingState')
            self.setCurrentHeatingCoolingState(climatizationStatus.climatisationState)
            self.charTargetHeatingCoolingState = servThermostat.configure_char('TargetHeatingCoolingState')

        if climatizationSettings is not None and climatizationSettings.targetTemperature_C.enabled:
            climatizationSettings.targetTemperature_C.addObserver(self.onTargetTemperatureChange,
                                                                  AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            self.charTargetTemperature = servThermostat.configure_char('TargetTemperature')
            self.charTargetTemperature.set_value(climatizationSettings.targetTemperature_C.value)

        # Set constant to celsius
        # TODO: We can enable conversion here later
        self.charTemperatureDisplayUnits = servThermostat.configure_char('TemperatureDisplayUnits')
        self.charTemperatureDisplayUnits.set_value(0)

    def setCurrentHeatingCoolingState(self, climatisationState):
        if self.charCurrentHeatingCoolingState is not None:
            if climatisationState.value == ClimatizationStatus.ClimatizationState.HEATING:
                self.charCurrentHeatingCoolingState.set_value(1)
            elif climatisationState.value == ClimatizationStatus.ClimatizationState.COOLING \
                    or climatisationState == ClimatizationStatus.ClimatizationState.VENTILATION:
                self.charCurrentHeatingCoolingState.set_value(2)
            elif climatisationState.value == ClimatizationStatus.ClimatizationState.OFF:
                self.charCurrentHeatingCoolingState.set_value(0)
            else:
                self.charCurrentHeatingCoolingState.set_value(0)
                LOG.warn('unsupported climatisationState: %s', climatisationState.value.value)

    def onClimatizationState(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.setCurrentHeatingCoolingState(element)
            LOG.debug('Climatization State Changed: %s', element.value.value)
        else:
            LOG.debug('Unsupported event %s', flags)

    def onTargetTemperatureChange(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.charTargetTemperature.set_value(element.value)
            LOG.debug('targetTemperature_C Changed: %d %%', element.value)
        else:
            LOG.debug('Unsupported event %s', flags)



import logging

import pyhap

from weconnect.elements.charging_status import ChargingStatus
from weconnect.elements.plug_status import PlugStatus
from weconnect.addressable import AddressableLeaf

LOG = logging.getLogger("VWsFriend")


class Charging(pyhap.accessory.Accessory):
    """Charging Assessory"""

    category = pyhap.const.CATEGORY_OUTLET

    def __init__(self, driver, displayName, chargingStatus, plugStatus=None):
        super().__init__(driver=driver, display_name=displayName)

        servOutlet = self.add_preload_service('Outlet')

        if chargingStatus is not None and chargingStatus.chargingState.enabled:
            chargingStatus.chargingState.addObserver(self.onChargingState, AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            self.charOn = servOutlet.configure_char('On')
            # print(self.charOn.properties)
            # self.charOn.override_properties(properties={'Format': 'bool', 'Permissions': ['pr', 'ev']})
            self.setOnState(chargingStatus.chargingState)

        # if chargingStatus is not None and chargingStatus.remainingChargingTimeToComplete_min.enabled:
        #     chargingStatus.remainingChargingTimeToComplete_min.addObserver(self.onRemainingChargingTimeToComplete,
        #                                                                    AddressableLeaf.ObserverEvent.VALUE_CHANGED)
        #     # Add Characteristic that is not planned for the service. This is still visible in other Apps than Apple Home
        #     servOutlet.add_characteristic(pyhap.loader.get_loader().get_char("RemainingDuration"))
        #     self.charRemainingDuration = servOutlet.configure_char('RemainingDuration')
        #     self.charRemainingDuration.set_value(chargingStatus.remainingChargingTimeToComplete_min.value * 60)

        if plugStatus is not None and plugStatus.plugConnectionState.enabled:
            plugStatus.plugConnectionState.addObserver(self.onplugConnectionStateChange, AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            self.charOutletInUse = servOutlet.configure_char('OutletInUse')
            self.setOutletInUse(plugStatus.plugConnectionState)

    def setOnState(self, chargingState):
        if self.charOn is not None:
            if chargingState.value == ChargingStatus.ChargingState.OFF \
                    or chargingState.value == ChargingStatus.ChargingState.READY_FOR_CHARGING:
                self.charOn.set_value(0)
            elif chargingState.value == ChargingStatus.ChargingState.CHARGING:
                self.charOn.set_value(1)
            elif chargingState.value == ChargingStatus.ChargingState.ERROR:
                self.charOn.set_value(0)
            else:
                self.charOn.set_value(0)
                LOG.warn('unsupported chargingState: %s', chargingState.value.value)

    def setOutletInUse(self, plugConnectionState):
        if self.charOutletInUse is not None:
            if plugConnectionState.value == PlugStatus.PlugConnectionState.CONNECTED:
                self.charOutletInUse.set_value(True)
            elif plugConnectionState.value == PlugStatus.PlugConnectionState.DISCONNECTED:
                self.charOutletInUse.set_value(False)
            else:
                self.charOutletInUse.set_value(False)
                LOG.warn('unsupported chargingState: %s', plugConnectionState.value.value)

    def onChargingState(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.setOnState(element)
            LOG.debug('Charging State Changed: %s', element.value.value)
        else:
            LOG.debug('Unsupported event %s', flags)

    def onplugConnectionStateChange(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.setOutletInUse(element)
            LOG.debug('Plug connection state Changed: %s', element.value.value)
        else:
            LOG.debug('Unsupported event %s', flags)

    def onRemainingChargingTimeToComplete(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.charRemainingDuration.set_value(element.value * 60)
            LOG.debug('RemainingChargingTimeToComplete Changed: %d %%', element.value)
        else:
            LOG.debug('Unsupported event %s', flags)



from sqlalchemy import and_

from vwsfriend.model.climatization import Climatization

from weconnect.addressable import AddressableLeaf


class ClimatizationAgent():
    def __init__(self, session, vehicle):
        self.session = session
        self.vehicle = vehicle
        self.charge = session.query(Climatization).filter(and_(Climatization.vehicle == vehicle,
                                                               Climatization.carCapturedTimestamp.isnot(None))) \
                                                  .order_by(Climatization.carCapturedTimestamp.desc()).first()

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            if 'climatisationStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['climatisationStatus'].enabled:
                self.vehicle.weConnectVehicle.statuses['climatisationStatus'].carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampChange,
                                                                                                               AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                               onUpdateComplete=True)
                self.__onCarCapturedTimestampChange(None, None)

    def __onCarCapturedTimestampChange(self, element, flags):
        chargeStatus = self.vehicle.weConnectVehicle.statuses['climatisationStatus']
        current_remainingClimatisationTime_min = None
        current_climatisationState = None
        if chargeStatus.remainingClimatisationTime_min.enabled:
            current_remainingClimatisationTime_min = chargeStatus.remainingClimatisationTime_min.value
        if chargeStatus.climatisationState.enabled:
            current_climatisationState = chargeStatus.climatisationState.value

        if self.charge is None or (self.charge.carCapturedTimestamp != chargeStatus.carCapturedTimestamp.value and (
                self.charge.remainingClimatisationTime_min != current_remainingClimatisationTime_min
                or self.charge.climatisationState != current_climatisationState)):

            self.charge = Climatization(self.vehicle, chargeStatus.carCapturedTimestamp.value, current_remainingClimatisationTime_min,
                                        current_climatisationState)
            self.session.add(self.charge)

    def commit(self):
        pass



from sqlalchemy import Column, String, Enum, Boolean
from sqlalchemy.orm import relationship

from weconnect.addressable import AddressableLeaf
from weconnect.elements.range_status import RangeStatus

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator


class Vehicle(Base):
    __tablename__ = 'vehicles'
    vin = Column(String(17), primary_key=True)
    model = Column(String(256))
    nickname = Column(String(256))
    carType = Column(Enum(RangeStatus.CarType, length=63))
    online = Column(Boolean)
    lastUpdate = Column(DatetimeDecorator)
    lastChange = Column(DatetimeDecorator)
    settings = relationship("VehicleSettings", back_populates="vehicle", uselist=False)

    weConnectVehicle = None

    def __init__(self, vin):
        self.vin = vin

    def connect(self, weConnectVehicle):
        self.weConnectVehicle = weConnectVehicle
        self.weConnectVehicle.model.addObserver(self.__onModelChange, AddressableLeaf.ObserverEvent.VALUE_CHANGED)
        if self.weConnectVehicle.model.enabled and self.model != self.weConnectVehicle.model.value:
            self.model = self.weConnectVehicle.model.value
        self.weConnectVehicle.nickname.addObserver(self.__onNicknameChange, AddressableLeaf.ObserverEvent.VALUE_CHANGED)
        if self.weConnectVehicle.nickname.enabled and self.nickname != self.weConnectVehicle.nickname.value:
            self.nickname = self.weConnectVehicle.nickname.value
        if 'rangeStatus' in self.weConnectVehicle.statuses and self.weConnectVehicle.statuses['rangeStatus'].enabled:
            self.weConnectVehicle.statuses['rangeStatus'].carType.addObserver(self.__onCarTypeChange, AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            if self.weConnectVehicle.statuses['rangeStatus'].carType.enabled:
                self.carType = self.weConnectVehicle.statuses['rangeStatus'].carType.value
            elif self.carType is None:
                self.carType = RangeStatus.CarType.UNKNOWN

    def __onModelChange(self, element, flags):
        if self.model != element.value:
            self.model = element.value

    def __onNicknameChange(self, element, flags):
        if self.nickname != element.value:
            self.nickname = element.value

    def __onCarTypeChange(self, element, flags):
        if self.carType != element.value:
            self.carType = element.value



import re
from collections import namedtuple

from flask import Blueprint, render_template, current_app, abort, redirect, url_for, request
from flask_wtf import FlaskForm
from wtforms import FieldList, FormField, StringField, IntegerField, SubmitField, HiddenField
from wtforms.validators import DataRequired, NumberRange, Optional, Length, Regexp, ValidationError

from weconnect.elements.range_status import RangeStatus

from vwsfriend.model.vehicle_settings import VehicleSettings
from vwsfriend.model.vehicle import Vehicle
from vwsfriend.agents.abrp.abrp_agent import ABRPAgent


bp = Blueprint('settings', __name__, url_prefix='/settings')


class VehicleSettingsForm(FlaskForm):
    primary_capacity = IntegerField('Primary Capacity', validators=[DataRequired(), NumberRange(min=1, max=500)])
    primary_wltp_range = IntegerField('Primary WLTP range in km', validators=[DataRequired(), NumberRange(min=1, max=1000)])
    secondary_capacity = IntegerField('Secondary Capacity', validators=[Optional(), NumberRange(min=1, max=500)], filters=[lambda x: x or None])
    secondary_wltp_range = IntegerField('Secondary WLTP Range', validators=[Optional(), NumberRange(min=1, max=1000)], filters=[lambda x: x or None])
    save = SubmitField('Save')


class ElectricVehicleSettingsForm(VehicleSettingsForm):
    primary_capacity = IntegerField('Usable Battery Capacity (net) in kWh', validators=[DataRequired(), NumberRange(min=1, max=500)])
    primary_wltp_range = IntegerField('WLTP range in km', validators=[DataRequired(), NumberRange(min=1, max=1000)])
    secondary_capacity = HiddenField('Secondary Capacity', validators=[Optional(), NumberRange(min=1, max=500)], filters=[lambda x: x or None])
    secondary_wltp_range = HiddenField('Secondary WLTP Range', validators=[Optional(), NumberRange(min=1, max=1000)], filters=[lambda x: x or None])
    save = SubmitField('Save')


class HybridVehicleSettingsForm(VehicleSettingsForm):
    primary_capacity = IntegerField('Gasoline tank size in liters', validators=[DataRequired(), NumberRange(min=1, max=500)])
    primary_wltp_range = IntegerField('WLTP range in km gasoline only', validators=[DataRequired(), NumberRange(min=1, max=1000)])
    secondary_capacity = IntegerField('Usable Battery Capacity (net) in kWh', validators=[Optional(), NumberRange(min=1, max=500)],
                                      filters=[lambda x: x or None])
    secondary_wltp_range = IntegerField('WLTP range in km electic only', validators=[Optional(), NumberRange(min=1, max=1000)], filters=[lambda x: x or None])
    save = SubmitField('Save')


class ABRPSettingsForm(FlaskForm):
    class Account(FlaskForm):
        account_name = StringField('Account Name', validators=[Optional(), Length(min=1, max=255)])
        account_token = StringField('User Token', validators=[Optional(), Regexp(regex=r'[0-9a-f\-]{100}', message='token format is not correct')])
        delete = SubmitField('Delete')

    AccountEntry = namedtuple('Account', ('account_name', 'account_token'))

    accounts = FieldList(FormField(Account, 'ABRP Account'), min_entries=0, max_entries=10)
    save = SubmitField('Save')
    addAccount = SubmitField(label='Add Account')

    def validate_accounts(form, field):
        for data in field.data:
            if len(data['account_name']) < 1:
                raise ValidationError('Account name must not be empty')
            if not re.match(r'[0-9a-f\-]{10,100}', data['account_token']):
                raise ValidationError('Token is no real ABRP user token')


ABRPSettingsEntry = namedtuple('ABRPSettings', ('accounts'))


@bp.route('/vehicle/<vin>', methods=['GET', 'POST'])
def vehicle(vin):
    if vin not in current_app.weConnect.vehicles:
        abort(404, f"Vehicle with VIN {vin} doesn't exist.")
    foundVehicle = current_app.weConnect.vehicles[vin]

    return render_template('settings/vehicle.html', vehicle=foundVehicle, connector=current_app.connector)


@bp.route('/vehicle/database/<vin>', methods=['GET', 'POST'])
def vehicleDBParameters(vin):
    if vin not in current_app.weConnect.vehicles:
        abort(404, f"Vehicle with VIN {vin} doesn't exist.")
    foundVehicle = current_app.weConnect.vehicles[vin]
    dbVehicle = current_app.session.query(Vehicle).filter(Vehicle.vin == vin).first()
    if dbVehicle.settings is None:
        dbVehicle.settings = VehicleSettings(vehicle=dbVehicle)
        current_app.session.commit()

    if dbVehicle.carType == RangeStatus.CarType.ELECTRIC:
        form = ElectricVehicleSettingsForm()
    elif dbVehicle.carType == RangeStatus.CarType.HYBRID:
        form = HybridVehicleSettingsForm()
    else:
        form = VehicleSettingsForm()

    if request.method == "GET":
        form.primary_capacity.data = dbVehicle.settings.primary_capacity
        form.primary_wltp_range.data = dbVehicle.settings.primary_wltp_range
        form.secondary_capacity.data = dbVehicle.settings.secondary_capacity
        form.secondary_wltp_range.data = dbVehicle.settings.secondary_wltp_range
    elif form.validate_on_submit():
        dbVehicle.settings.primary_capacity = form.primary_capacity.data
        dbVehicle.settings.primary_wltp_range = form.primary_wltp_range.data
        dbVehicle.settings.secondary_capacity = form.secondary_capacity.data
        dbVehicle.settings.secondary_wltp_range = form.secondary_wltp_range.data
        current_app.session.commit()

        return redirect(url_for("settings.vehicle", vin=vin))

    return render_template('settings/vehicledbparameters.html', vehicle=foundVehicle, form=form, connector=current_app.connector)


@bp.route('/vehicle/abrp/<vin>', methods=['GET', 'POST'])  # noqa: C901
def vehicleABRPSettings(vin):  # noqa: C901
    if vin not in current_app.weConnect.vehicles:
        abort(404, f"Vehicle with VIN {vin} doesn't exist.")
    foundVehicle = current_app.weConnect.vehicles[vin]
    agent = None
    for connectorAgent in current_app.connector.agents[vin]:
        if isinstance(connectorAgent, ABRPAgent):
            agent = connectorAgent
    if agent is None:
        abort(404, f"ABRP was not enabled for vehicle with VIN {vin}")

    tokens = agent.userTokens
    data = ABRPSettingsEntry([ABRPSettingsForm.AccountEntry(x[0], x[1]) for x in tokens])
    form = ABRPSettingsForm(obj=data)

    validate = True

    if form.addAccount.data:
        form.accounts.append_entry()
        validate = False
    else:
        keepList = list()
        for account in form.accounts.entries:
            if not account.delete.data:
                keepList.append(account.data)
            else:
                validate = False
        for _ in range(len(form.accounts.entries)):
            form.accounts.pop_entry()
        for keep in keepList:
            form.accounts.append_entry(keep)

    if validate and form.validate_on_submit():
        newTokens = list()
        for entry in form.accounts.entries:
            newTokens.append((entry.account_name.data, entry.account_token.data))
        agent.userTokens = newTokens
        return redirect(url_for("settings.vehicle", vin=vin))

    return render_template('settings/abrpsettings.html', vehicle=foundVehicle, form=form, connector=current_app.connector)



from sqlalchemy import Column, Integer, String, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator


class Online(Base):
    __tablename__ = 'onlinestates'
    __table_args__ = (
        UniqueConstraint('vehicle_vin', 'onlineTime'),
    )
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    onlineTime = Column(DatetimeDecorator)
    offlineTime = Column(DatetimeDecorator)
    vehicle = relationship("Vehicle")

    def __init__(self, vehicle, onlineTime, offlineTime):
        self.vehicle = vehicle
        self.onlineTime = onlineTime
        self.offlineTime = offlineTime



import logging
import json
import requests

from weconnect.elements.vehicle import Vehicle
from weconnect.elements.charging_status import ChargingStatus

from vwsfriend.__version import __version__

LOG = logging.getLogger("VWsFriend")

API_BASE_URL = 'https://api.iternio.com/1/'
VWSFRIEND_IDENTIFIER = '6225724a-65fb-4d4c-9ac5-d7dff2b78c1d'

HEADER = {'accept': 'application/json',
          'user-agent': f'VWsFriend ({__version__})',
          'accept-language': 'en-en',
          'Authorization': f'APIKEY {VWSFRIEND_IDENTIFIER}'}


class ABRPAgent():
    def __init__(self, weConnectVehicle: Vehicle, tokenfile):
        self.weConnectVehicle = weConnectVehicle
        self.__session = requests.Session()
        self.__session.headers = HEADER
        self.__userTokens = list()

        self.telemetryData = dict()
        self.tokenfile = tokenfile
        try:
            self.readTokens(tokenfile)
        except FileNotFoundError:
            pass

    @property
    def userTokens(self):
        return self.__userTokens

    @userTokens.setter
    def userTokens(self, userTokens):
        self.__userTokens = userTokens
        self.persistTokens(self.tokenfile)

    def persistTokens(self, tokenfile):
        if tokenfile:
            try:
                with open(tokenfile, 'w') as file:
                    json.dump(self.__userTokens, fp=file)
                LOG.info('Writing tokenfile %s', tokenfile)
            except ValueError as err:  # pragma: no cover
                LOG.info('Could not write abrp tokenfile %s (%s)', tokenfile, err)

    def readTokens(self, tokenfile):
        with open(tokenfile, 'r') as file:
            tokens = json.load(fp=file)
            self.__userTokens = tokens
        LOG.info('Reading abrp tokenfile %s', tokenfile)

    def updateTelemetry(self):
        for account, token in self.__userTokens:
            params = {'token': token}
            data = {'tlm': self.telemetryData}
            response = self.__session.post(API_BASE_URL + 'tlm/send', params=params, json=data)
            if response.status_code != requests.codes['ok']:
                LOG.error(
                    f'ABRP send telemetry for vehicle {self.weConnectVehicle.vin.value} for account {account} failed with status code {response.status_code}')
            else:
                data = response.json()
                if 'status' in data:
                    if data['status'] != 'ok':
                        LOG.error(f'ABRP send telemetry for vehicle {self.weConnectVehicle.vin.value} for account {account} failed')
                    if 'missing' in data:
                        LOG.info(f'ABRP send telemetry for vehicle {self.weConnectVehicle.vin.value} for account {account}: {data["missing"]}')
                else:
                    LOG.error(f'ABRP send telemetry for vehicle {self.weConnectVehicle.vin.value} for account {account} returned unexpected data')

    def commit(self):  # noqa: C901
        if 'batteryStatus' in self.weConnectVehicle.statuses:
            batteryStatus = self.weConnectVehicle.statuses['batteryStatus']
            if batteryStatus.carCapturedTimestamp.enabled and batteryStatus.carCapturedTimestamp.value is not None:
                self.telemetryData['utc'] = batteryStatus.carCapturedTimestamp.value.timestamp()
            if batteryStatus.currentSOC_pct.enabled and batteryStatus.currentSOC_pct.value is not None:
                self.telemetryData['soc'] = batteryStatus.currentSOC_pct.value

        if 'parkingPosition' in self.weConnectVehicle.statuses:
            parkingPosition = self.weConnectVehicle.statuses['parkingPosition']
            if parkingPosition.latitude.enabled and parkingPosition.latitude.value is not None \
                    and parkingPosition.longitude.enabled and parkingPosition.longitude.value is not None:
                self.telemetryData['lat'] = parkingPosition.latitude.value
                self.telemetryData['lon'] = parkingPosition.longitude.value
        if 'lat' in self.telemetryData:
            self.telemetryData['is_parked'] = True
        else:
            self.telemetryData['is_parked'] = False

        if 'maintenanceStatus' in self.weConnectVehicle.statuses and self.weConnectVehicle.statuses['maintenanceStatus'].enabled:
            maintenanceStatus = self.weConnectVehicle.statuses['maintenanceStatus']
            if maintenanceStatus.mileage_km.enabled and maintenanceStatus.mileage_km is not None:
                self.telemetryData['odometer'] = maintenanceStatus.mileage_km.value

        if 'chargingStatus' in self.weConnectVehicle.statuses and self.weConnectVehicle.statuses['chargingStatus'].enabled:
            chargingStatus = self.weConnectVehicle.statuses['chargingStatus']
            if chargingStatus.chargingState.enabled and chargingStatus.chargingState.value in [ChargingStatus.ChargingState.CHARGING]:
                self.telemetryData['is_charging'] = True
            else:
                self.telemetryData['is_charging'] = False

            if chargingStatus.chargePower_kW.enabled and chargingStatus.chargePower_kW.value is not None:
                self.telemetryData['power'] = chargingStatus.chargePower_kW.value * -1

        self.updateTelemetry()
        self.telemetryData = dict()



import logging

import pyhap

from .climatization import Climatization
from .battery import Battery
from .charging import Charging

from vwsfriend.__version import __version__

LOG = logging.getLogger("VWsFriend")


class VWsFriendBridge(pyhap.accessory.Bridge):
    """VWsfriend Bridge"""

    def __init__(self, weConnect, driver, displayName='VWsFriend'):
        super().__init__(driver=driver, display_name=displayName, )

        self.set_info_service(__version__, "Till Steinbach", "VWsFriend", None)

        self.weConnect = weConnect
        self.driver = driver

    def update(self):
        for vehicle in self.weConnect.vehicles.values():
            manufacturer = 'Volkswagen'
            nickname = vehicle.nickname.value
            model = vehicle.nickname.value
            vin = vehicle.vin.value
            # Climatization
            if 'climatisationStatus' in vehicle.statuses:
                climatizationStatus = vehicle.statuses['climatisationStatus']

                if 'climatisationSettings' in vehicle.statuses:
                    climatizationSettings = vehicle.statuses['climatisationSettings']
                else:
                    climatizationSettings = None
                climatizationAccessory = Climatization(driver=self.driver, displayName=f'{nickname} Climatization',
                                                       climatizationStatus=climatizationStatus,
                                                       climatizationSettings=climatizationSettings)
                climatizationAccessory.set_info_service(manufacturer=manufacturer,
                                                        model=model,
                                                        serial_number=vin)
                self.add_accessory(climatizationAccessory)

            if 'batteryStatus' in vehicle.statuses:
                batteryStatus = vehicle.statuses['batteryStatus']

                if 'chargingStatus' in vehicle.statuses:
                    chargingStatus = vehicle.statuses['chargingStatus']
                else:
                    chargingStatus = None

                batteryAccessory = Battery(driver=self.driver, displayName=f'{nickname} Battery',
                                           batteryStatus=batteryStatus,
                                           chargingStatus=chargingStatus)
                batteryAccessory.set_info_service(manufacturer=manufacturer,
                                                  model=model,
                                                  serial_number=vin)
                self.add_accessory(batteryAccessory)

            if 'chargingStatus' in vehicle.statuses:
                chargingStatus = vehicle.statuses['chargingStatus']

                if 'plugStatus' in vehicle.statuses:
                    plugStatus = vehicle.statuses['plugStatus']
                else:
                    plugStatus = None

                chargingAccessory = Charging(driver=self.driver, displayName=f'{nickname} Charging',
                                             chargingStatus=chargingStatus,
                                             plugStatus=plugStatus)
                chargingAccessory.set_info_service(manufacturer=manufacturer,
                                                   model=model,
                                                   serial_number=vin)
                self.add_accessory(chargingAccessory)



from .battery import Battery
from .charge import Charge
from .charging_session import ChargingSession
from .climatization import Climatization
from .location import Location
from .online import Online
from .range import Range
from .refuel_session import RefuelSession
from .trip import Trip
from .vehicle_settings import VehicleSettings
from .vehicle import Vehicle
from .charger import Charger, Operator



import logging

import pyhap

from weconnect.elements.charging_status import ChargingStatus
from weconnect.addressable import AddressableLeaf

LOG = logging.getLogger("VWsFriend")


class Battery(pyhap.accessory.Accessory):
    """Battery Assessory"""

    category = pyhap.const.CATEGORY_OTHER

    def __init__(self, driver, displayName, batteryStatus, chargingStatus=None):
        super().__init__(driver=driver, display_name=displayName)

        servBattery = self.add_preload_service('BatteryService')

        if batteryStatus.currentSOC_pct.enabled:
            batteryStatus.currentSOC_pct.addObserver(self.onCurrentSOCChange,
                                                     AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            self.charBatteryLevel = servBattery.configure_char('BatteryLevel')
            self.charBatteryLevel.set_value(batteryStatus.currentSOC_pct.value)
            self.charStatusLowBattery = servBattery.configure_char('StatusLowBattery')
            self.setStatusLowBattery(batteryStatus.currentSOC_pct)

        if chargingStatus is not None and chargingStatus.chargingState.enabled:
            chargingStatus.chargingState.addObserver(self.onChargingState, AddressableLeaf.ObserverEvent.VALUE_CHANGED)
            self.charChargingState = servBattery.configure_char('ChargingState')
            self.setChargingState(chargingStatus.chargingState)

    def setStatusLowBattery(self, currentSOC_pct):
        if self.charStatusLowBattery is not None:
            if currentSOC_pct.value > 10:
                self.charStatusLowBattery.set_value(0)
            else:
                self.charStatusLowBattery.set_value(1)

    def setChargingState(self, chargingState):
        if self.charChargingState is not None:
            if chargingState.value == ChargingStatus.ChargingState.OFF \
                    or chargingState.value == ChargingStatus.ChargingState.READY_FOR_CHARGING:
                self.charChargingState.set_value(0)
            elif chargingState.value == ChargingStatus.ChargingState.CHARGING:
                self.charChargingState.set_value(1)
            elif chargingState.value == ChargingStatus.ChargingState.ERROR:
                self.charChargingState.set_value(2)
            else:
                self.charChargingState.set_value(2)
                LOG.warn('unsupported chargingState: %s', chargingState.value.value)

    def onCurrentSOCChange(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.charBatteryLevel.set_value(element.value)
            self.setStatusLowBattery(element)
            LOG.debug('Battery SoC Changed: %d %%', element.value)
        else:
            LOG.debug('Unsupported event %s', flags)

    def onChargingState(self, element, flags):
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED:
            self.setChargingState(element)
            LOG.debug('Charging State Changed: %s', element.value.value)
        else:
            LOG.debug('Unsupported event %s', flags)



from sqlalchemy import Column, Integer, BigInteger, Float, String, ForeignKey
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator


class RefuelSession(Base):
    __tablename__ = 'refuel_sessions'
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    vehicle = relationship("Vehicle")

    date = Column(DatetimeDecorator)
    startSOC_pct = Column(Integer)
    endSOC_pct = Column(Integer)
    mileage_km = Column(Integer)
    position_latitude = Column(Float)
    position_longitude = Column(Float)
    location_id = Column(BigInteger, ForeignKey('locations.osm_id'))
    location = relationship("Location")

    def __init__(self, vehicle, date, startSOC_pct, endSOC_pct, mileage_km, position_latitude, position_longitude, location):
        self.vehicle = vehicle
        self.date = date
        self.startSOC_pct = startSOC_pct
        self.endSOC_pct = endSOC_pct
        self.mileage_km = mileage_km
        self.position_latitude = position_latitude
        self.position_longitude = position_longitude
        self.location = location



from io import BytesIO
from flask import Blueprint, render_template, current_app, abort, send_file

bp = Blueprint('status', __name__, url_prefix='/status')


@bp.route('/vehicles', methods=['GET'])
def vehicles():
    vehicles = current_app.weConnect.vehicles.values()
    return render_template('status/vehicles.html', vehicles=vehicles, connector=current_app.connector)


@bp.route('/vehicles/<string:vin>-status.png', methods=['GET'])
def vehicleStatusImg(vin):
    vehicles = current_app.weConnect.vehicles
    if vin not in vehicles:
        abort(404, f"Vehicle with VIN {vin} doesn't exist.")
    pictures = vehicles[vin].pictures
    if 'status' not in pictures:
        abort(404, "Status picture doesn't exist.")
    else:
        img_io = BytesIO()
        pictures['status'].value.save(img_io, 'PNG')
        img_io.seek(0)
    return send_file(img_io, mimetype='image/png')


@bp.route('/vehicles/<string:vin>-car.png', methods=['GET'])
def vehicleImg(vin):
    vehicles = current_app.weConnect.vehicles
    if vin not in vehicles:
        abort(404, f"Vehicle with VIN {vin} doesn't exist.")
    pictures = vehicles[vin].pictures
    if 'car' not in pictures:
        abort(404, "Status picture doesn't exist.")
    else:
        img_io = BytesIO()
        pictures['car'].value.save(img_io, 'PNG')
        img_io.seek(0)
    return send_file(img_io, mimetype='image/png')



import logging

from vwsfriend.model.refuel_session import RefuelSession
from vwsfriend.util.location_util import locationFromLatLon

from weconnect.addressable import AddressableLeaf
from weconnect.elements.range_status import RangeStatus

LOG = logging.getLogger("VWsFriend")


class RefuelAgent():
    def __init__(self, session, vehicle):
        self.session = session
        self.vehicle = vehicle
        self.primary_currentSOC_pct = None

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            if 'rangeStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['rangeStatus'].enabled:
                self.vehicle.weConnectVehicle.statuses['rangeStatus'].carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampChange,
                                                                                                       AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                       onUpdateComplete=True)
                self.__onCarCapturedTimestampChange(None, None)

    def __onCarCapturedTimestampChange(self, element, flags):
        rangeStatus = self.vehicle.weConnectVehicle.statuses['rangeStatus']
        if self.vehicle.carType in [RangeStatus.CarType.HYBRID] and rangeStatus.primaryEngine.currentSOC_pct.enabled:
            current_primary_currentSOC_pct = rangeStatus.primaryEngine.currentSOC_pct.value

            mileage_km = None
            if 'maintenanceStatus' in self.vehicle.weConnectVehicle.statuses:
                maintenanceStatus = self.vehicle.weConnectVehicle.statuses['maintenanceStatus']
                if maintenanceStatus.mileage_km.enabled:
                    mileage_km = maintenanceStatus.mileage_km.value

            position_latitude = None
            position_longitude = None
            location = None
            if 'parkingPosition' in self.vehicle.weConnectVehicle.statuses:
                parkingPosition = self.vehicle.weConnectVehicle.statuses['parkingPosition']
                if parkingPosition.latitude.enabled and parkingPosition.latitude.value is not None \
                        and parkingPosition.longitude.enabled and parkingPosition.longitude.value is not None:
                    position_latitude = parkingPosition.latitude.value
                    position_longitude = parkingPosition.longitude.value
                    location = locationFromLatLon(self.session, parkingPosition.latitude.value, parkingPosition.longitude.value)

            # Refuel event took place
            if self.primary_currentSOC_pct is not None and current_primary_currentSOC_pct > self.primary_currentSOC_pct:
                LOG.info('Vehicle %s refueled from %d percent to %d percent', self.vehicle.vin, self.primary_currentSOC_pct, current_primary_currentSOC_pct)
                self.range = RefuelSession(self.vehicle, element.value, self.primary_currentSOC_pct, current_primary_currentSOC_pct, mileage_km,
                                           position_latitude, position_longitude, location)
                self.session.add(self.range)
                self.primary_currentSOC_pct = current_primary_currentSOC_pct
            # SoC decreased, normal usage
            elif self.primary_currentSOC_pct is None or current_primary_currentSOC_pct < self.primary_currentSOC_pct:
                self.primary_currentSOC_pct = current_primary_currentSOC_pct

    def commit(self):
        pass



import json

from sqlalchemy import Column, BigInteger, String, Float

from vwsfriend.model.base import Base


class Location(Base):
    __tablename__ = 'locations'
    osm_id = Column(BigInteger, primary_key=True)
    osm_type = Column(String)
    latitude = Column(Float)
    longitude = Column(Float)
    display_name = Column(String)
    name = Column(String)
    amenity = Column(String)
    house_number = Column(String)
    road = Column(String)
    neighbourhood = Column(String)
    city = Column(String)
    postcode = Column(String)
    county = Column(String)
    country = Column(String)
    state = Column(String)
    state_district = Column(String)
    raw = Column(String)

    def __init__(self, jsonDict):  # noqa: C901
        if 'osm_id' in jsonDict and jsonDict['osm_id'] is not None:
            self.osm_id = int(jsonDict['osm_id'])

        if 'osm_type' in jsonDict and jsonDict['osm_type'] is not None:
            self.osm_type = jsonDict['osm_type']

        if 'lat' in jsonDict and jsonDict['lat'] is not None:
            self.latitude = jsonDict['lat']

        if 'lon' in jsonDict and jsonDict['lon'] is not None:
            self.longitude = jsonDict['lon']

        if 'display_name' in jsonDict and jsonDict['display_name'] is not None:
            self.display_name = jsonDict['display_name']

        if 'address' in jsonDict and jsonDict['address'] is not None:
            address = jsonDict['address']

            if 'amenity' in address and address['amenity'] is not None:
                self.amenity = address['amenity']

            for attribute in ['house_number', 'street_number']:
                if attribute in address and address[attribute] is not None:
                    self.house_number = address[attribute]
                    break

            for attribute in ['road', 'footway', 'street', 'street_name', 'residential', 'path', 'pedestrian', 'road_reference', 'road_reference_intl',
                              'square', 'place']:
                if attribute in address and address[attribute] is not None:
                    self.road = address[attribute]
                    break

            for attribute in ['neighbourhood', 'suburb', 'city_district', 'district', 'quarter', 'borough', 'city_block', 'residential', 'commercial',
                              'industrial', 'houses', 'subdistrict', 'subdivision', 'ward']:
                if attribute in address and address[attribute] is not None:
                    self.neighbourhood = address[attribute]
                    break

            for attribute in ['city', 'town', 'township']:
                if attribute in address and address[attribute] is not None:
                    self.city = address[attribute]
                    break

            for attribute in ['postcode', 'partial_postcode']:
                if attribute in address and address[attribute] is not None:
                    self.postcode = address[attribute]
                    break

            for attribute in ['county', 'county_code', 'department']:
                if attribute in address and address[attribute] is not None:
                    self.county = address[attribute]
                    break

            for attribute in ['country', 'country_name']:
                if attribute in address and address[attribute] is not None:
                    self.country = address[attribute]
                    break

            for attribute in ['state', 'province', 'state_code']:
                if attribute in address and address[attribute] is not None:
                    self.state = address[attribute]
                    break

            if 'state_district' in jsonDict and jsonDict['state_district'] is not None:
                self.state_district = jsonDict['state_district']

        if 'namedetails' in jsonDict and jsonDict['namedetails'] is not None:
            namedetails = jsonDict['namedetails']
            for attribute in ['name', 'alt_name']:
                if attribute in namedetails and namedetails[attribute] is not None:
                    self.name = namedetails[attribute]
                    break

        self.raw = json.dumps(jsonDict)

    def __str__(self):  # noqa: C901
        returnString = f'OSM ID: {self.osm_id}'
        if self.osm_type is not None:
            returnString += f'\nOSM Type: {self.osm_type}'
        if self.latitude is not None:
            returnString += f'\nLatitude: {self.latitude}'
        if self.longitude is not None:
            returnString += f'\nLongitude: {self.longitude}'
        if self.name is not None:
            returnString += f'\nName: {self.name}'
        if self.display_name is not None:
            returnString += f'\nDisplayName: {self.display_name}'
        if self.amenity is not None:
            returnString += f'\nAmenity: {self.amenity}'
        if self.road is not None:
            returnString += f'\nRoad: {self.road}'
        if self.house_number is not None:
            returnString += f'\nHouse Number: {self.house_number}'
        if self.neighbourhood is not None:
            returnString += f'\nNighbourhood: {self.neighbourhood}'
        if self.city is not None:
            returnString += f'\nCity: {self.city}'
        if self.postcode is not None:
            returnString += f'\nPostcode: {self.postcode}'
        if self.county is not None:
            returnString += f'\nCounty: {self.county}'
        if self.country is not None:
            returnString += f'\nCountry: {self.country}'
        if self.state is not None:
            returnString += f'\nState: {self.state}'
        if self.state_district is not None:
            returnString += f'\nState District: {self.state_district}'
        return returnString



import os
import sys
import re
import argparse
from datetime import datetime, timedelta, timezone
import logging
import time
import tempfile
import netrc
import getpass

import threading

from pyhap.accessory_driver import AccessoryDriver

from weconnect import weconnect
from weconnect.__version import __version__ as __weconnect_version__

from vwsfriend.ui.vwsfriend_ui import VWsFriendUI
from vwsfriend.homekit.bridge import VWsFriendBridge
from vwsfriend.agent_connector import AgentConnector

from .__version import __version__


LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
DEFAULT_LOG_LEVEL = "ERROR"

LOG = logging.getLogger("VWsFriend")


class NumberRangeArgument:

    def __init__(self, imin=None, imax=None):
        self.imin = imin
        self.imax = imax

    def __call__(self, arg):
        try:
            value = int(arg)
        except ValueError as e:
            raise self.exception() from e
        if (self.imin is not None and value < self.imin) or (self.imax is not None and value > self.imax):
            raise self.exception()
        return value

    def exception(self):
        if self.imin is not None and self.imax is not None:
            return argparse.ArgumentTypeError(f'Must be a number from {self.imin} to {self.imax}')
        if self.imin is not None:
            return argparse.ArgumentTypeError(f'Must be a number not smaller than {self.imin}')
        if self.imax is not None:
            return argparse.ArgumentTypeError(f'Must be number not larger than {self.imax}')

        return argparse.ArgumentTypeError('Must be a number')


def main():  # noqa: C901 pylint: disable=too-many-branches, too-many-statements, too-many-locals
    parser = argparse.ArgumentParser(
        prog='vwsfriend',
        description='TBD')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version} (using WeConnect-python {weversion})'.format(version=__version__, weversion=__weconnect_version__))
    weConnectGroup = parser.add_argument_group('WeConnect')
    weConnectGroup.add_argument('-u', '--username', help='Username of Volkswagen id', required=False)
    weConnectGroup.add_argument('-p', '--password', help='Password of Volkswagen id', required=False)
    defaultNetRc = os.path.join(os.path.expanduser("~"), ".netrc")
    weConnectGroup.add_argument('--netrc', help=f'File in netrc syntax providing login (default: {defaultNetRc}).'
                                ' Netrc is only used when username and password are not provided  as arguments',
                                default=None, required=False)
    weConnectGroup.add_argument('-i', '--interval', help='Query interval in seconds',
                                type=NumberRangeArgument(imin=300), required=False, default=300)
    defaultTemp = os.path.join(tempfile.gettempdir(), 'weconnect.token')
    weConnectGroup.add_argument('--tokenfile', help=f'file to store token (default: {defaultTemp})', default=defaultTemp)
    weConnectGroup.add_argument('--no-token-storage', dest='noTokenStorage', help='Do not store token on filesystem (this'
                                ' will cause a new login for every invokation!)', action='store_true')

    parser.add_argument('-v', '--verbose', action="append_const", const=-1,)
    parser.add_argument('--config-dir', dest='configDir', help='directory to store configuration files (default: ./)', default='./')
    parser.add_argument('--demo', help='folder containing demo scenario, see README for more information')
    dbGroup = parser.add_argument_group('Database & visualization')

    dbGroup.add_argument('--with-database', dest='withDatabase', help='Connect VWsFriend to database for visualization', action='store_true')
    dbGroup.add_argument('--database-url', dest='dbUrl', help='Database to connect to', default='sqlite:///vwsfrienddevel.db')

    abrpGroup = parser.add_argument_group('ABRP: A better route planner')
    abrpGroup.add_argument('--with-abrp', dest='withABRP', help='Connect VWsFriend to ABRP (you need to add userTokens in the UI!)', action='store_true')

    homekitGroup = parser.add_argument_group('Homekit')
    homekitGroup.add_argument('--with-homekit', dest='withHomekit', help='Provide Apple Homekit functionality', action='store_true')

    args = parser.parse_args()

    logLevel = LOG_LEVELS.index(DEFAULT_LOG_LEVEL)
    for adjustment in args.verbose or ():
        logLevel = min(len(LOG_LEVELS) - 1, max(logLevel + adjustment, 0))

    logging.basicConfig(level=LOG_LEVELS[logLevel])
    logging.getLogger("pyhap").setLevel(level="CRITICAL")
    # logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

    LOG.info('vwsfriend %s (using WeConnect-python %s)', __version__, __weconnect_version__)

    username = None
    password = None

    if args.username is not None and args.password is not None:
        username = args.username
        password = args.password
    else:
        if args.netrc is not None:
            netRcFilename = args.netrc
        else:
            netRcFilename = defaultNetRc
        try:
            secrets = netrc.netrc(file=args.netrc)
            username, _, password = secrets.authenticators("volkswagen.de")
        except TypeError:
            if not args.username:
                LOG.error('volkswagen.de entry was not found in %s netrc-file. Create it or provide at least a username'
                          ' with --username', netRcFilename)
                sys.exit(1)
            username = args.username
            password = getpass.getpass()
        except FileNotFoundError:
            if not args.username:
                LOG.error('%s netrc-file was not found. Create it or provide at least a username with --username',
                          netRcFilename)
                sys.exit(1)
            username = args.username
            password = getpass.getpass()
    tokenfile = None
    if not args.noTokenStorage:
        tokenfile = args.tokenfile

    try:
        weConnect = weconnect.WeConnect(username=username, password=password, tokenfile=tokenfile,
                                        updateAfterLogin=False, loginOnInit=(args.demo is None))

        connector = AgentConnector(weConnect=weConnect, dbUrl=args.dbUrl, interval=args.interval, withDB=args.withDatabase, withABRP=args.withABRP,
                                   configDir=args.configDir)

        if args.withHomekit:
            LOG.info('Starting up Homekit')
            # Start the accessory on port 51826
            driver = AccessoryDriver(pincode=b'123-45-678', persist_file=f'{args.configDir}/accessory.state', port=51826)
            bridge = VWsFriendBridge(driver=driver, weConnect=weConnect)
            driver.add_accessory(bridge)
            weConnectBridgeInitialized = False

            # Start it!
            hapThread = threading.Thread(target=driver.start)
            hapThread.start()

        ui = VWsFriendUI(weConnect=weConnect, connector=connector, dbUrl=args.dbUrl)
        ui.run()

        if args.demo is not None:
            utcDemoStart = datetime.utcnow().replace(tzinfo=timezone.utc, microsecond=0)
            for file in sorted(os.listdir(args.demo)):
                fileNameRegex = r'(?P<number>\d+)_(?P<delay>\d+)s(_(?P<stage>[^\.]+))?.cache.json'
                match = re.search(fileNameRegex, file)
                if match is not None:
                    time.sleep(int(match.groupdict()['delay']))
                    stageFilePath = f'{args.demo}/{file}'
                    with open(stageFilePath, mode='r', encoding='utf8') as fp:
                        cacheString = fp.read()
                        cacheString = re.sub(r'demodate\((?P<offset>[+-]?\d+)\)',
                                             lambda m: str(utcDemoStart + timedelta(seconds=int(m.groupdict()['offset']))).replace('+00:00', 'Z'), cacheString)
                        cacheString = re.sub(r'now\((?P<offset>[+-]?\d+)\)',
                                             lambda m: str(datetime.now() + timedelta(seconds=int(m.groupdict()['offset']))), cacheString)
                        weConnect.fillCacheFromJsonString(cacheString, maxAge=2147483647)
                        if args.withHomekit and not weConnectBridgeInitialized:
                            weConnectBridgeInitialized = True
                            bridge.update()
                        weConnect.update(updateCapabilities=True)
                        connector.commit()
                        if match.groupdict()['stage'] is not None:
                            LOG.info('Stage %s completed', match.groupdict()['stage'])
                        else:
                            LOG.info('Stage completed')
            LOG.info('Demo completed')
        else:
            while True:
                try:
                    LOG.info('Updating data from WeConnect')
                    weConnect.update(updateCapabilities=True, updatePictures=True, force=True)
                    connector.commit()
                    if args.withHomekit and not weConnectBridgeInitialized:
                        weConnectBridgeInitialized = True
                        bridge.update()
                except weconnect.RetrievalError:
                    LOG.error('Retrieval error during update. Will try again after configured interval of %ds', args.interval)
                time.sleep(args.interval)

    except weconnect.AuthentificationError as e:
        LOG.critical('There was a problem when authenticating with WeConnect: %s', e)
    except weconnect.APICompatibilityError as e:
        LOG.critical('There was a problem when communicating with WeConnect.'
                     ' If this problem persists please open a bug report: %s', e)



from sqlalchemy import and_

from vwsfriend.model.battery import Battery

from weconnect.addressable import AddressableLeaf


class BatteryAgent():
    def __init__(self, session, vehicle):
        self.session = session
        self.vehicle = vehicle
        self.battery = session.query(Battery).filter(and_(Battery.vehicle == vehicle,
                                                          Battery.carCapturedTimestamp.isnot(None))).order_by(Battery.carCapturedTimestamp.desc()).first()

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            if 'batteryStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['batteryStatus'].enabled:
                self.vehicle.weConnectVehicle.statuses['batteryStatus'].carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampChange,
                                                                                                         AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                         onUpdateComplete=True)
                self.__onCarCapturedTimestampChange(None, None)

    def __onCarCapturedTimestampChange(self, element, flags):
        batteryStatus = self.vehicle.weConnectVehicle.statuses['batteryStatus']
        current_currentSOC_pct = batteryStatus.currentSOC_pct.value
        current_cruisingRangeElectric_km = batteryStatus.cruisingRangeElectric_km.value

        if self.battery is None or (self.battery.carCapturedTimestamp != batteryStatus.carCapturedTimestamp.value and (
                self.battery.currentSOC_pct != current_currentSOC_pct
                or self.battery.cruisingRangeElectric_km != current_cruisingRangeElectric_km)):

            self.battery = Battery(self.vehicle, batteryStatus.carCapturedTimestamp.value, current_currentSOC_pct, current_cruisingRangeElectric_km)
            self.session.add(self.battery)

    def commit(self):
        pass



from sqlalchemy import Column, Integer, String, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator


class Range(Base):
    __tablename__ = 'ranges'
    __table_args__ = (
        UniqueConstraint('vehicle_vin', 'carCapturedTimestamp'),
    )
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    carCapturedTimestamp = Column(DatetimeDecorator(timezone=True), nullable=False)
    vehicle = relationship("Vehicle")
    totalRange_km = Column(Integer)
    primary_currentSOC_pct = Column(Integer)
    primary_remainingRange_km = Column(Integer)
    secondary_currentSOC_pct = Column(Integer)
    secondary_remainingRange_km = Column(Integer)

    def __init__(self, vehicle, carCapturedTimestamp, totalRange_km, primary_currentSOC_pct, primary_remainingRange_km, secondary_currentSOC_pct,
                 secondary_remainingRange_km):
        self.vehicle = vehicle
        self.carCapturedTimestamp = carCapturedTimestamp
        self.totalRange_km = totalRange_km
        self.primary_currentSOC_pct = primary_currentSOC_pct
        self.primary_remainingRange_km = primary_remainingRange_km
        self.secondary_currentSOC_pct = secondary_currentSOC_pct
        self.secondary_remainingRange_km = secondary_remainingRange_km




from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base


class VehicleSettings(Base):
    __tablename__ = 'vehicle_settings'
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'), primary_key=True)
    vehicle = relationship("Vehicle", back_populates="settings")

    primary_capacity = Column(Integer)
    primary_wltp_range = Column(Integer)
    secondary_capacity = Column(Integer)
    secondary_wltp_range = Column(Integer)

    def __init__(self, vehicle):
        self.vehicle = vehicle



from sqlalchemy import and_

from vwsfriend.model.charge import Charge
from vwsfriend.model.charging_session import ChargingSession, ACDC
from vwsfriend.util.location_util import locationFromLatLon, chargerFromLatLon

from weconnect.addressable import AddressableLeaf
from weconnect.elements.charging_status import ChargingStatus
from weconnect.elements.plug_status import PlugStatus


class ChargeAgent():
    def __init__(self, session, vehicle):
        self.session = session
        self.vehicle = vehicle
        self.charge = session.query(Charge).filter(and_(Charge.vehicle == vehicle, Charge.carCapturedTimestamp.isnot(None))
                                                   ).order_by(Charge.carCapturedTimestamp.desc()).first()
        self.chargingSession = None
        self.previousChargingSession = None

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            if 'chargingStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['chargingStatus'].enabled:
                self.vehicle.weConnectVehicle.statuses['chargingStatus'].carCapturedTimestamp.addObserver(self.__onChargingStatusCarCapturedTimestampChange,
                                                                                                          AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                          onUpdateComplete=True)
                self.__onChargingStatusCarCapturedTimestampChange(None, None)

                self.vehicle.weConnectVehicle.statuses['chargingStatus'].chargingState.addObserver(self.__onChargingStateChange,
                                                                                                   AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                   onUpdateComplete=True)
                self.vehicle.weConnectVehicle.statuses['chargingStatus'].chargePower_kW.addObserver(self.__onChargePowerChange,
                                                                                                    AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                    onUpdateComplete=True)

            if 'plugStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['plugStatus'].enabled:
                self.vehicle.weConnectVehicle.statuses['plugStatus'].plugConnectionState.addObserver(self.__onPlugConnectionStateChange,
                                                                                                     AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                     onUpdateComplete=True)
                self.vehicle.weConnectVehicle.statuses['plugStatus'].plugLockState.addObserver(self.__onPlugLockStateChange,
                                                                                               AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                               onUpdateComplete=True)

    def __onChargingStatusCarCapturedTimestampChange(self, element, flags):
        chargeStatus = self.vehicle.weConnectVehicle.statuses['chargingStatus']
        current_remainingChargingTimeToComplete_min = None
        current_chargingState = None
        current_chargeMode = None
        current_chargePower_kW = None
        current_chargeRate_kmph = None
        if chargeStatus.remainingChargingTimeToComplete_min.enabled:
            current_remainingChargingTimeToComplete_min = chargeStatus.remainingChargingTimeToComplete_min.value
        if chargeStatus.chargingState.enabled:
            current_chargingState = chargeStatus.chargingState.value
        if chargeStatus.chargeMode.enabled:
            current_chargeMode = chargeStatus.chargeMode.value
        if chargeStatus.chargePower_kW.enabled:
            current_chargePower_kW = chargeStatus.chargePower_kW.value
        if chargeStatus.chargeRate_kmph.enabled:
            current_chargeRate_kmph = chargeStatus.chargeRate_kmph.value

        if self.charge is None or (self.charge.carCapturedTimestamp != chargeStatus.carCapturedTimestamp.value and (
                self.charge.remainingChargingTimeToComplete_min != current_remainingChargingTimeToComplete_min
                or self.charge.chargingState != current_chargingState
                or self.charge.chargeMode != current_chargeMode
                or self.charge.chargePower_kW != current_chargePower_kW
                or self.charge.chargeRate_kmph != current_chargeRate_kmph)):

            self.charge = Charge(self.vehicle, chargeStatus.carCapturedTimestamp.value, current_remainingChargingTimeToComplete_min, current_chargingState,
                                 current_chargeMode, current_chargePower_kW, current_chargeRate_kmph)
            self.session.add(self.charge)

    def __onChargingStateChange(self, element, flags):  # noqa: C901
        chargeStatus = self.vehicle.weConnectVehicle.statuses['chargingStatus']
        if element.value == ChargingStatus.ChargingState.CHARGING:
            if self.chargingSession is None or self.chargingSession.isClosed():
                self.previousChargingSession = self.chargingSession
                self.chargingSession = ChargingSession(vehicle=self.vehicle)
                self.session.add(self.chargingSession)
            if not self.chargingSession.wasStarted():
                self.chargingSession.started = chargeStatus.carCapturedTimestamp.value
            # also write start SoC
            if self.chargingSession is not None and self.chargingSession.startSOC_pct is None \
                    and 'batteryStatus' in self.vehicle.weConnectVehicle.statuses:
                batteryStatus = self.vehicle.weConnectVehicle.statuses['batteryStatus']
                if batteryStatus.enabled and batteryStatus.currentSOC_pct.enabled:
                    self.chargingSession.startSOC_pct = batteryStatus.currentSOC_pct.value

            # also write position if available
            self.updatePosition()

            # also write milage if available
            self.updateMileage()
        elif element.value in [ChargingStatus.ChargingState.OFF, ChargingStatus.ChargingState.READY_FOR_CHARGING]:
            if self.chargingSession is not None and self.chargingSession.isChargingState():
                self.chargingSession.ended = chargeStatus.carCapturedTimestamp.value

                if self.chargingSession.maximumChargePower_kW is not None:
                    if self.chargingSession.maximumChargePower_kW > 11:
                        self.chargingSession.acdc = ACDC.DC
                    else:
                        self.chargingSession.acdc = ACDC.AC
                # also write end SoC
                if self.chargingSession is not None and self.chargingSession.endSOC_pct is None \
                        and 'batteryStatus' in self.vehicle.weConnectVehicle.statuses:
                    batteryStatus = self.vehicle.weConnectVehicle.statuses['batteryStatus']
                    if batteryStatus.enabled and batteryStatus.currentSOC_pct.enabled:
                        self.chargingSession.endSOC_pct = batteryStatus.currentSOC_pct.value

                # also write position if available
                self.updatePosition()

                # also write milage if available
                self.updateMileage()

    def __onPlugConnectionStateChange(self, element, flags):
        plugStatus = self.vehicle.weConnectVehicle.statuses['plugStatus']
        if element.value == PlugStatus.PlugConnectionState.CONNECTED:
            if self.chargingSession is None or self.chargingSession.isClosed():
                self.previousChargingSession = self.chargingSession
                self.chargingSession = ChargingSession(vehicle=self.vehicle)
                self.session.add(self.chargingSession)
            if self.chargingSession.connected is None:
                self.chargingSession.connected = plugStatus.carCapturedTimestamp.value
            # also write position if available
            self.updatePosition()
            # also write milage if available
            self.updateMileage()
        elif element.value == PlugStatus.PlugConnectionState.DISCONNECTED:
            if self.chargingSession is not None and self.chargingSession.isConnectedState():
                self.chargingSession.disconnected = plugStatus.carCapturedTimestamp.value
            # also write position if available
            self.updatePosition()
            # also write milage if available
            self.updateMileage()

    def __onPlugLockStateChange(self, element, flags):
        plugStatus = self.vehicle.weConnectVehicle.statuses['plugStatus']
        if element.value == PlugStatus.PlugLockState.LOCKED:
            if self.chargingSession is None or self.chargingSession.isClosed():
                self.previousChargingSession = self.chargingSession
                self.chargingSession = ChargingSession(vehicle=self.vehicle)
                self.session.add(self.chargingSession)
            if self.chargingSession.locked is None:
                self.chargingSession.locked = plugStatus.carCapturedTimestamp.value
            # also write position if available
            self.updatePosition()
            # also write milage if available
            self.updateMileage()
        elif element.value == PlugStatus.PlugLockState.UNLOCKED:
            if self.chargingSession is not None and self.chargingSession.isLockedState():
                self.chargingSession.unlocked = plugStatus.carCapturedTimestamp.value
            # also write position if available
            self.updatePosition()
            # also write milage if available
            self.updateMileage()

    def __onChargePowerChange(self, element, flags):
        if self.chargingSession.isChargingState() and self.chargingSession is not None \
                and (self.chargingSession.maximumChargePower_kW is None or element.value > self.chargingSession.maximumChargePower_kW):
            self.chargingSession.maximumChargePower_kW = element.value

    def updatePosition(self):
        if 'parkingPosition' in self.vehicle.weConnectVehicle.statuses:
            parkingPosition = self.vehicle.weConnectVehicle.statuses['parkingPosition']
            if self.chargingSession is not None and parkingPosition.latitude.enabled and parkingPosition.latitude.value is not None \
                    and parkingPosition.longitude.enabled and parkingPosition.longitude.value is not None:
                self.chargingSession.position_latitude = parkingPosition.latitude.value
                self.chargingSession.position_longitude = parkingPosition.longitude.value
                if self.chargingSession.location is None:
                    self.chargingSession.location = locationFromLatLon(self.session, parkingPosition.latitude.value, parkingPosition.longitude.value)
                if self.chargingSession.charger is None:
                    self.chargingSession.charger = chargerFromLatLon(weConnect=self.vehicle.weConnectVehicle.weConnect, session=self.session,
                                                                     latitude=parkingPosition.latitude.value, longitude=parkingPosition.longitude.value,
                                                                     searchRadius=100)

    def updateMileage(self):
        if 'maintenanceStatus' in self.vehicle.weConnectVehicle.statuses:
            maintenanceStatus = self.vehicle.weConnectVehicle.statuses['maintenanceStatus']
            if self.chargingSession is not None and maintenanceStatus.mileage_km.enabled and maintenanceStatus.mileage_km.value is not None:
                self.chargingSession.mileage_km = maintenanceStatus.mileage_km.value

    def commit(self):
        pass



from sqlalchemy import Column, Integer, String, Enum, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator

from weconnect.elements.climatization_status import ClimatizationStatus


class Climatization(Base):
    __tablename__ = 'climatization'
    __table_args__ = (
        UniqueConstraint('vehicle_vin', 'carCapturedTimestamp'),
    )
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    carCapturedTimestamp = Column(DatetimeDecorator(timezone=True), nullable=False)
    vehicle = relationship("Vehicle")
    remainingClimatisationTime_min = Column(Integer)
    climatisationState = Column(Enum(ClimatizationStatus.ClimatizationState, length=63))

    def __init__(self, vehicle, carCapturedTimestamp, remainingClimatisationTime_min, climatisationState):
        self.vehicle = vehicle
        self.carCapturedTimestamp = carCapturedTimestamp
        self.remainingClimatisationTime_min = remainingClimatisationTime_min
        self.climatisationState = climatisationState



import logging
from datetime import datetime, timezone

from sqlalchemy import and_

from vwsfriend.model.trip import Trip
from vwsfriend.util.location_util import locationFromLatLon

from weconnect.addressable import AddressableLeaf, AddressableAttribute

LOG = logging.getLogger("VWsFriend")


class TripAgent():
    def __init__(self, session, vehicle):
        self.session = session
        self.vehicle = vehicle

        self.trip = session.query(Trip).filter(and_(Trip.vehicle == vehicle, Trip.startDate.isnot(None))).order_by(Trip.startDate.desc()).first()
        if self.trip is not None:
            if self.trip.endDate is not None:
                self.lastParkingPositionTimestamp = self.trip.endDate
                self.lastParkingPositionLatitude = self.trip.destination_position_latitude
                self.lastParkingPositionLongitude = self.trip.destination_position_longitude
            else:
                LOG.info(f'Vehicle {self.vehicle.vin} has still an open trip during startup, closing it now')
            self.trip = None
        else:
            self.lastParkingPositionTimestamp = None
            self.lastParkingPositionLatitude = None
            self.lastParkingPositionLongitude = None

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            if 'parkingPosition' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['parkingPosition'].enabled:
                self.vehicle.weConnectVehicle.statuses['parkingPosition'].carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampChange,
                                                                                                           AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                           onUpdateComplete=True)
                self.__onCarCapturedTimestampChange(self.vehicle.weConnectVehicle.statuses['parkingPosition'].carCapturedTimestamp, None)
                self.vehicle.weConnectVehicle.statuses['parkingPosition'].carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampDisabled,
                                                                                                           AddressableLeaf.ObserverEvent.DISABLED,
                                                                                                           onUpdateComplete=True)
                LOG.info(f'Vehicle {self.vehicle.vin} provides a parkingPosition and thus allows to record trips')
            else:
                self.vehicle.weConnectVehicle.statuses.addObserver(self.__onStatusesChange,
                                                                   AddressableLeaf.ObserverEvent.ENABLED,
                                                                   onUpdateComplete=True)

    def __onStatusesChange(self, element, flags):
        if isinstance(element, AddressableAttribute) and element.getGlobalAddress().endswith('parkingPosition/carCapturedTimestamp'):
            # only add if not in list of observers
            if self.__onCarCapturedTimestampChange not in element.getObservers(flags=AddressableLeaf.ObserverEvent.VALUE_CHANGED, onUpdateComplete=True):
                element.addObserver(self.__onCarCapturedTimestampChange,
                                    AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                    onUpdateComplete=True)
                element.addObserver(self.__onCarCapturedTimestampDisabled,
                                    AddressableLeaf.ObserverEvent.DISABLED,
                                    onUpdateComplete=True)
                LOG.info(f'Vehicle {self.vehicle.vin} provides a parkingPosition and thus allows to record trips')
                self.vehicle.weConnectVehicle.statuses.removeObserver(self.__onStatusesChange)
                self.__onCarCapturedTimestampChange(element, flags)

    def __onCarCapturedTimestampDisabled(self, element, flags):
        if self.trip is not None:
            LOG.info(f'Vehicle {self.vehicle.vin} removed a parkingPosition but there was an open trip, closing it now')
            self.trip = None
        self.trip = Trip(self.vehicle, datetime.utcnow().replace(tzinfo=timezone.utc, microsecond=0), self.lastParkingPositionLatitude,
                         self.lastParkingPositionLongitude, None, None)
        self.trip.start_location = locationFromLatLon(self.session, self.lastParkingPositionLatitude, self.lastParkingPositionLongitude)

        if 'maintenanceStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['maintenanceStatus'].enabled:
            maintenanceStatus = self.vehicle.weConnectVehicle.statuses['maintenanceStatus']
            if maintenanceStatus.mileage_km.enabled and maintenanceStatus.mileage_km is not None:
                self.trip.start_mileage_km = maintenanceStatus.mileage_km.value

        self.session.add(self.trip)
        LOG.info(f'Vehicle {self.vehicle.vin} started a trip')

    def __onCarCapturedTimestampChange(self, element, flags):
        parkingPosition = self.vehicle.weConnectVehicle.statuses['parkingPosition']
        if parkingPosition.carCapturedTimestamp.enabled and parkingPosition.carCapturedTimestamp.value is not None:
            self.lastParkingPositionTimestamp = parkingPosition.carCapturedTimestamp.value
        if parkingPosition.latitude.enabled and parkingPosition.latitude.value is not None \
                and parkingPosition.longitude.enabled and parkingPosition.longitude.value is not None:
            self.lastParkingPositionLatitude = parkingPosition.latitude.value
            self.lastParkingPositionLongitude = parkingPosition.longitude.value
        if self.trip is not None:
            if parkingPosition.carCapturedTimestamp.enabled and parkingPosition.carCapturedTimestamp.value is not None:
                self.trip.endDate = parkingPosition.carCapturedTimestamp.value
            if parkingPosition.latitude.enabled and parkingPosition.latitude.value is not None \
                    and parkingPosition.longitude.enabled and parkingPosition.longitude.value is not None:
                self.trip.destination_position_latitude = parkingPosition.latitude.value
                self.trip.destination_position_longitude = parkingPosition.longitude.value
                self.trip.destination_location = locationFromLatLon(self.session, parkingPosition.latitude.value, parkingPosition.longitude.value)

            if 'maintenanceStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['maintenanceStatus'].enabled:
                maintenanceStatus = self.vehicle.weConnectVehicle.statuses['maintenanceStatus']
                if maintenanceStatus.mileage_km.enabled and maintenanceStatus.mileage_km is not None:
                    self.trip.end_mileage_km = maintenanceStatus.mileage_km.value

            self.trip = None

            LOG.info(f'Vehicle {self.vehicle.vin} ended a trip')
        else:
            if flags is not None:
                LOG.info(f'Vehicle {self.vehicle.vin} provides a parking position, but no trip was started (this is ok during startup)')

    def commit(self):
        pass



import enum

from sqlalchemy import Column, Integer, BigInteger, Float, String, Enum, ForeignKey
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator

from weconnect.elements.charging_settings import ChargingSettings


class ACDC(enum.Enum,):
    AC = enum.auto()
    DC = enum.auto()
    UNKNOWN = enum.auto()


class ChargingSession(Base):
    __tablename__ = 'charging_sessions'
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    vehicle = relationship("Vehicle")

    connected = Column(DatetimeDecorator)
    locked = Column(DatetimeDecorator)
    started = Column(DatetimeDecorator)
    ended = Column(DatetimeDecorator)
    unlocked = Column(DatetimeDecorator)
    disconnected = Column(DatetimeDecorator)
    maxChargeCurrentACSetting = Column(Enum(ChargingSettings.MaximumChargeCurrent, length=63))
    targetSOCSetting_pct = Column(Integer)
    maximumChargePower_kW = Column(Integer)
    acdc = Column(Enum(ACDC, length=63))
    startSOC_pct = Column(Integer)
    endSOC_pct = Column(Integer)
    mileage_km = Column(Integer)
    position_latitude = Column(Float)
    position_longitude = Column(Float)
    location_id = Column(BigInteger, ForeignKey('locations.osm_id'))
    location = relationship("Location")
    charger_id = Column(String, ForeignKey('chargers.id'))
    charger = relationship("Charger")

    def __init__(self, vehicle):
        self.vehicle = vehicle

    def isConnectedState(self):
        return self.connected is not None and self.disconnected is None

    def isLockedState(self):
        return self.locked is not None and self.unlocked is None

    def isChargingState(self):
        return self.started is not None and self.ended is None

    def isClosed(self):
        return self.ended is not None or self.unlocked is not None or self.disconnected is not None

    def wasStarted(self):
        return self.started is not None

    def wasConnected(self):
        return self.connected is not None

    def wasLocked(self):
        return self.locked is not None

    def wasEnded(self):
        return self.ended is not None

    def wasDisconnected(self):
        return self.disconnected is not None

    def wasUnlocked(self):
        return self.unlocked is not None



from sqlalchemy import Column, String, Float, Integer, ForeignKey
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base


class Charger(Base):
    __tablename__ = 'chargers'
    id = Column(String, primary_key=True)
    name = Column(String)
    latitude = Column(Float)
    longitude = Column(Float)
    address = Column(String)
    max_power = Column(Float)
    num_spots = Column(Integer)
    operator_id = Column(String, ForeignKey('operators.id'))
    operator = relationship("Operator")

    def __init__(self, id):
        self.id = id


class Operator(Base):
    __tablename__ = 'operators'
    id = Column(String, primary_key=True)
    name = Column(String)
    phone = Column(String)

    def __init__(self, id, name, phone):
        self.id = id
        self.name = name
        self.phone = phone



from sqlalchemy import and_

from vwsfriend.model.range import Range

from weconnect.addressable import AddressableLeaf


class RangeAgent():
    def __init__(self, session, vehicle):
        self.session = session
        self.vehicle = vehicle
        self.range = session.query(Range).filter(and_(Range.vehicle == vehicle,
                                                      Range.carCapturedTimestamp.isnot(None))).order_by(Range.carCapturedTimestamp.desc()).first()

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            if 'rangeStatus' in self.vehicle.weConnectVehicle.statuses and self.vehicle.weConnectVehicle.statuses['rangeStatus'].enabled:
                self.vehicle.weConnectVehicle.statuses['rangeStatus'].carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampChange,
                                                                                                       AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                                                                       onUpdateComplete=True)
                self.__onCarCapturedTimestampChange(None, None)

    def __onCarCapturedTimestampChange(self, element, flags):
        rangeStatus = self.vehicle.weConnectVehicle.statuses['rangeStatus']
        current_totalRange_km = rangeStatus.totalRange_km.value
        current_primary_currentSOC_pct = None
        current_primary_remainingRange_km = None
        if rangeStatus.primaryEngine.enabled:
            current_primary_currentSOC_pct = rangeStatus.primaryEngine.currentSOC_pct.value
            current_primary_remainingRange_km = rangeStatus.primaryEngine.remainingRange_km.value
        current_secondary_currentSOC_pct = None
        current_secondary_remainingRange_km = None
        if rangeStatus.secondaryEngine.enabled:
            current_secondary_currentSOC_pct = rangeStatus.secondaryEngine.currentSOC_pct.value
            current_secondary_remainingRange_km = rangeStatus.secondaryEngine.remainingRange_km.value

        if self.range is None or (rangeStatus.carCapturedTimestamp.value is not None
                                  and self.range.carCapturedTimestamp != rangeStatus.carCapturedTimestamp.value
                                  and (self.range.totalRange_km != current_totalRange_km
                                       or self.range.primary_currentSOC_pct != current_primary_currentSOC_pct
                                       or self.range.primary_remainingRange_km != current_primary_remainingRange_km
                                       or self.range.secondary_currentSOC_pct != current_secondary_currentSOC_pct
                                       or self.range.secondary_remainingRange_km != current_secondary_remainingRange_km)):

            self.range = Range(self.vehicle, rangeStatus.carCapturedTimestamp.value, current_totalRange_km, current_primary_currentSOC_pct,
                               current_primary_remainingRange_km, current_secondary_currentSOC_pct, current_secondary_remainingRange_km)
            self.session.add(self.range)
            self.session.flush()

    def commit(self):
        pass



from vwsfriend.model.location import Location
from vwsfriend.model.charger import Charger, Operator
import requests

from weconnect.errors import RetrievalError


def locationFromLatLon(session, latitude, longitude):
    query = {
        'lat': latitude,
        'lon': longitude,
        'namedetails': 1,
        'format': 'json'
    }
    headers = {
        'User-Agent': 'VWsFriend'
    }

    response = requests.get('https://nominatim.openstreetmap.org/reverse', params=query, headers=headers)
    if response.status_code == requests.codes['ok']:
        location = Location(jsonDict=response.json())
        return session.merge(location)
    return None


def chargerFromLatLon(weConnect, session, latitude, longitude, searchRadius):
    try:
        chargers = sorted(weConnect.getChargingStations(latitude, longitude, searchRadius=searchRadius).values(), key=lambda station: station.distance.value)
        if len(chargers) > 0:
            charger = Charger(id=chargers[0].id.value)
            if chargers[0].name.enabled:
                charger.name = chargers[0].name.value
            if chargers[0].latitude.enabled:
                charger.latitude = chargers[0].latitude.value
            if chargers[0].longitude.enabled:
                charger.longitude = chargers[0].longitude.value
            if chargers[0].address.enabled:
                charger.address = str(chargers[0].address)
            if chargers[0].chargingPower.enabled:
                charger.max_power = chargers[0].chargingPower.value
            if chargers[0].chargingSpots.enabled:
                charger.num_spots = len(chargers[0].chargingSpots)

            charger.operator = Operator(id=chargers[0].operator.id.value, name=chargers[0].operator.name.value, phone=chargers[0].operator.phoneNumber.value)

            return session.merge(charger)
    except RetrievalError:
        pass
    return None



from enum import Enum, auto
from datetime import datetime, timezone, timedelta
import logging

from sqlalchemy import and_

from vwsfriend.model.online import Online

from weconnect.addressable import AddressableLeaf

LOG = logging.getLogger("VWsFriend")


class StateAgent():
    def __init__(self, session, vehicle, updateInterval):
        self.session = session
        self.vehicle = vehicle
        self.onlineTimeout = (updateInterval * 2) + 30
        self.offlineTimeout = (updateInterval * 2) + 30
        # The offline timeout must be at least 10 minutes plus 30 seconds as this is the update frequency of data for some cars.
        if self.offlineTimeout < 630:
            self.offlineTimeout = 630
        self.onlineState = None
        self.online = session.query(Online).filter(and_(Online.vehicle == vehicle, Online.onlineTime.isnot(None))).order_by(Online.onlineTime.desc()).first()
        # If the last record in the database is completed we are not online right now
        if self.online is None or self.online.offlineTime is not None:
            self.online = None
            self.onlineState = StateAgent.OnlineState.OFFLINE
        else:
            self.onlineState = StateAgent.OnlineState.ONLINE
            LOG.warning(f'Vehicle {vehicle.vin} is still online in database. Looks like the session was closed when the vehicle was still online!')
        self.lastCarCapturedTimestamp = None
        self.earliestCarCapturedTimestampInInterval = None

        # register for updates:
        if self.vehicle.weConnectVehicle is not None:
            for status in self.vehicle.weConnectVehicle.statuses.values():
                status.carCapturedTimestamp.addObserver(self.__onCarCapturedTimestampChange, AddressableLeaf.ObserverEvent.VALUE_CHANGED, onUpdateComplete=True)
                self.__onCarCapturedTimestampChange(element=status.carCapturedTimestamp, flags=None)

    def __onCarCapturedTimestampChange(self, element, flags):
        if element.enabled and (self.lastCarCapturedTimestamp is None or self.lastCarCapturedTimestamp < element.value):
            self.lastCarCapturedTimestamp = element.value
        if self.onlineState == StateAgent.OnlineState.OFFLINE:
            if element.enabled and (self.earliestCarCapturedTimestampInInterval is None or self.earliestCarCapturedTimestampInInterval > element.value) \
                    and (element.value + timedelta(seconds=self.onlineTimeout)) > datetime.utcnow().replace(tzinfo=timezone.utc):
                self.earliestCarCapturedTimestampInInterval = element.value

    def checkOnlineOffline(self):
        if self.onlineState == StateAgent.OnlineState.ONLINE:
            if self.online is not None and (self.lastCarCapturedTimestamp + timedelta(seconds=self.offlineTimeout)) \
                    < datetime.utcnow().replace(tzinfo=timezone.utc):
                LOG.info(f'Vehicle {self.vehicle.vin} went offline')
                self.onlineState = StateAgent.OnlineState.OFFLINE
                self.vehicle.online = False
                self.online.offlineTime = self.lastCarCapturedTimestamp
                self.online = None
                self.lastCarCapturedTimestamp = None
            else:
                if self.vehicle.lastChange is None or (self.lastCarCapturedTimestamp > self.vehicle.lastChange.replace(tzinfo=timezone.utc)):
                    self.vehicle.lastChange = self.lastCarCapturedTimestamp
        else:
            # When online now but now record add the record
            if self.online is None and self.earliestCarCapturedTimestampInInterval is not None:
                LOG.info(f'Vehicle {self.vehicle.vin} went online')
                self.onlineState = StateAgent.OnlineState.ONLINE
                self.vehicle.online = True
                if self.vehicle.lastChange is None or (self.lastCarCapturedTimestamp > self.vehicle.lastChange.replace(tzinfo=timezone.utc)):
                    self.vehicle.lastChange = self.lastCarCapturedTimestamp
                self.online = Online(self.vehicle, onlineTime=self.earliestCarCapturedTimestampInInterval, offlineTime=None)
                self.session.add(self.online)
                self.earliestCarCapturedTimestampInInterval = None

    def commit(self):
        self.checkOnlineOffline()
        self.vehicle.lastUpdate = datetime.utcnow().replace(tzinfo=timezone.utc)

    class OnlineState(Enum):
        ONLINE = auto()
        OFFLINE = auto()



from datetime import datetime, timezone
import sqlalchemy


class DatetimeDecorator(sqlalchemy.types.TypeDecorator):
    impl = sqlalchemy.types.DateTime
    cache_ok = True

    LOCAL_TIMEZONE = datetime.utcnow().astimezone().tzinfo

    def process_bind_param(self, value: datetime, dialect):
        if value is None:
            return value

        if value.tzinfo is None:
            value = value.astimezone(self.LOCAL_TIMEZONE)

        return value.astimezone(timezone.utc)

    def process_result_value(self, value, dialect):
        if value is None:
            return value

        if value.tzinfo is None:
            return value.replace(tzinfo=timezone.utc)

        return value.astimezone(timezone.utc)



from sqlalchemy import Column, Integer, String, Enum, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator

from weconnect.elements.charging_status import ChargingStatus


class Charge(Base):
    __tablename__ = 'charges'
    __table_args__ = (
        UniqueConstraint('vehicle_vin', 'carCapturedTimestamp'),
    )
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    carCapturedTimestamp = Column(DatetimeDecorator(timezone=True), nullable=False)
    vehicle = relationship("Vehicle")
    remainingChargingTimeToComplete_min = Column(Integer)
    chargingState = Column(Enum(ChargingStatus.ChargingState, length=63))
    chargeMode = Column(Enum(ChargingStatus.ChargeMode, length=63))
    chargePower_kW = Column(Integer)
    chargeRate_kmph = Column(Integer)

    def __init__(self, vehicle, carCapturedTimestamp, remainingChargingTimeToComplete_min, chargingState, chargeMode, chargePower_kW, chargeRate_kmph):
        self.vehicle = vehicle
        self.carCapturedTimestamp = carCapturedTimestamp
        self.remainingChargingTimeToComplete_min = remainingChargingTimeToComplete_min
        self.chargingState = chargingState
        self.chargeMode = chargeMode
        self.chargePower_kW = chargePower_kW
        self.chargeRate_kmph = chargeRate_kmph



from sqlalchemy import Column, Integer, BigInteger, Float, String, ForeignKey
from sqlalchemy.orm import relationship

from vwsfriend.model.base import Base
from vwsfriend.model.datetime_decorator import DatetimeDecorator


class Trip(Base):
    __tablename__ = 'trips'
    id = Column(Integer, primary_key=True)
    vehicle_vin = Column(String, ForeignKey('vehicles.vin'))
    vehicle = relationship("Vehicle")

    startDate = Column(DatetimeDecorator)
    endDate = Column(DatetimeDecorator)
    start_position_latitude = Column(Float)
    start_position_longitude = Column(Float)
    start_location_id = Column(BigInteger, ForeignKey('locations.osm_id'))
    start_location = relationship("Location", foreign_keys=[start_location_id])
    destination_position_latitude = Column(Float)
    destination_position_longitude = Column(Float)
    destination_location_id = Column(BigInteger, ForeignKey('locations.osm_id'))
    destination_location = relationship("Location", foreign_keys=[destination_location_id])
    start_mileage_km = Column(Integer)
    end_mileage_km = Column(Integer)

    def __init__(self, vehicle, startDate, start_position_latitude, start_position_longitude, start_location, start_mileage_km):
        self.vehicle = vehicle
        self.startDate = startDate
        self.start_position_latitude = start_position_latitude
        self.start_position_longitude = start_position_longitude
        self.start_location = start_location
        self.start_mileage_km = start_mileage_km



import logging

from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from weconnect.elements import vehicle
from weconnect.addressable import AddressableLeaf
from weconnect.elements.range_status import RangeStatus

from vwsfriend.agents.range_agent import RangeAgent
from vwsfriend.agents.battery_agent import BatteryAgent
from vwsfriend.agents.charge_agent import ChargeAgent
from vwsfriend.agents.state_agent import StateAgent
from vwsfriend.agents.climatization_agent import ClimatizationAgent
from vwsfriend.agents.refuel_agent import RefuelAgent
from vwsfriend.agents.trip_agent import TripAgent
from vwsfriend.agents.abrp.abrp_agent import ABRPAgent
from vwsfriend.model.base import Base
from vwsfriend.model import Vehicle

LOG = logging.getLogger("VWsFriend")


class AgentConnector():
    def __init__(self, weConnect, dbUrl, interval, withDB=False, withABRP=False, configDir='./'):
        self.agents = {}

        if withDB:
            engine = create_engine(dbUrl)
            self.session = Session(engine)
            Base.metadata.create_all(engine)

            self.vehicles = self.session.query(Vehicle).all()
        self.withDB = withDB

        self.withABRP = withABRP

        self.interval = interval
        self.configDir = configDir

        weConnect.addObserver(self.onEnable, AddressableLeaf.ObserverEvent.ENABLED, onUpdateComplete=True)

    def onEnable(self, element, flags):
        if (flags & AddressableLeaf.ObserverEvent.ENABLED) and isinstance(element, vehicle.Vehicle):
            if element.vin not in self.agents:
                self.agents[element.vin.value] = []
            if self.withDB:
                foundVehicle = None
                for dbVehicle in self.vehicles:
                    if dbVehicle.vin == element.vin.value:
                        LOG.info('Found matching vehicle for vin %s in database', element.vin.value)
                        foundVehicle = dbVehicle
                        break
                if foundVehicle is None:
                    LOG.info('Found no matching vehicle for vin %s in database, will create a new one', element.vin.value)
                    foundVehicle = Vehicle(element.vin.value)
                    self.session.add(foundVehicle)
                    self.session.commit(foundVehicle)
                foundVehicle.connect(element)

                self.agents[element.vin.value].append(RangeAgent(self.session, foundVehicle))
                self.agents[element.vin.value].append(BatteryAgent(self.session, foundVehicle))
                self.agents[element.vin.value].append(ChargeAgent(self.session, foundVehicle))
                self.agents[element.vin.value].append(StateAgent(self.session, foundVehicle, updateInterval=self.interval))
                self.agents[element.vin.value].append(ClimatizationAgent(self.session, foundVehicle))
                self.agents[element.vin.value].append(RefuelAgent(self.session, foundVehicle))
                self.agents[element.vin.value].append(TripAgent(self.session, foundVehicle))
                if foundVehicle.carType == RangeStatus.CarType.UNKNOWN:
                    LOG.warning('Vehicle %s has an unkown carType, thus some features won\'t be available until the correct carType could be detected',
                                foundVehicle.vin)
            if self.withABRP:
                self.agents[element.vin.value].append(ABRPAgent(weConnectVehicle=element, tokenfile=f'{self.configDir}/{element.vin.value}-ABRP.token'))

    def commit(self):
        for vehicleAgents in self.agents.values():
            for agent in vehicleAgents:
                agent.commit()
        if self.withDB:
            self.session.commit()
